#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cstring>
#include <cstdlib>
#include <iomanip>
using namespace std;

const int nodeMappingInvLen = 18;
const int nodeMappingLen = 51;
char nodeMappingInv[nodeMappingInvLen] = {4, 5, 8, 10, 12, 13, 19, 24, 25, 27, 29, 34, 38, 39, 40, 42, 44, 50};
char nodeMapping[nodeMappingLen] = {-1, -1, -1, -1, 0, 1, -1, -1, 2, -1, 3, -1, 4, 5, -1, -1, -1, -1, -1, 6, -1, -1, -1, -1, 7, 8, -1, 9, -1, 10, -1, -1, -1, -1, 11, -1, -1, -1, 12, 13, 14, -1, 15, -1, 16, -1, -1, -1, -1, -1, 17};

typedef struct _periodData {
    int localPeriod;
    int localTime;
    int globalTime;
    vector<int> rssi;
    _periodData() {
        localPeriod = 0;
        localTime = 0;
        globalTime = 0;
        rssi.resize(nodeMappingInvLen);
        for (int i = 0; i < nodeMappingInvLen; i++) rssi[i] = 0;
    }
} periodData;

typedef struct _master_time {
    int localTime;
    int masterTime;
} master_time;

const int MAX_LINE_LEN = 256;
int main(int argc, char** argv) {

    string   fileName = argv[1];
    char     strIn[MAX_LINE_LEN];
    vector<master_time> timeVect;
    vector<master_time> timeVectComp;
    char*    str = NULL;
    string  line;

    // Processing
    // 1. Read stored ticks from master timer
    // 2. Compensate for resets (timer runnning again from begging) occuring in master timer
    //   - resets occured after 65295381, then master time is broadcasted every 20 seconds
    // 3. Load social data
    // 4. Compute corresponding master time for given local period from stored master timer

    //
    //  Read master time
    //
    fileName += ".time";
    ifstream finTime(fileName.c_str(), ifstream::in);
    cout << "Input file with master time: " << fileName << endl;
    if (finTime.is_open()) {
        master_time item;
        while (finTime.getline(strIn, MAX_LINE_LEN)) {
            line = strIn;
            if (line == "Received message...TIME") {
                finTime.getline(strIn, MAX_LINE_LEN);
                //Message: 0 My time: 127 Master time: 4034581.
                str = strstr(strIn, "My time: ");
                str += strlen("My time: ");
                item.localTime = atoi(str);
                str = strstr(strIn, "Master time: ");
                str += strlen("Master time: ");
                item.masterTime = atoi(str);

                // Stop when unused memory is found
                int stopValue = 4294967295;
                if (item.masterTime == stopValue) break;

                // Store data into vector
                timeVect.push_back(item);
            }
        }
    }
    else cout << "Fail to open input file with time" << endl;
    finTime.close();
    cout << "Total time entries readed: " << timeVect.size() << endl;


    if (timeVect.size() > 0) {
        // Find time where first reset occured
        // Find next time item when master time will be above 204820 (sufficiently after reset)
        bool    bResetFound = false;
        int     lastMaster = 0;
        int     timeCompensation = 0;
        master_time resetTime = {0, 0};
        for (vector<master_time>::iterator iter = timeVect.begin(); iter != timeVect.end(); iter++) {
            // We reached the reset period of master time
            if (lastMaster > iter->masterTime) {
                bResetFound = true;
                resetTime = *(--iter);
                iter++;
                lastMaster = 0;
            }
            if (!bResetFound) lastMaster = iter->masterTime;

            // We reached safe period when new timer is up and running
            if (bResetFound && (iter->masterTime > 1003540)) {
                bResetFound = false;
                //Message: 11926 My time: 61261215 Master time: 65295381.
                //Message: 11992 My time: 62763250 Master time: 1003540.
                // Compute time compenastion for master time to remove effect of reset and timer restart
                timeCompensation = (iter->localTime - resetTime.localTime) + (resetTime.masterTime - iter->masterTime);
            }

            iter->masterTime += timeCompensation;
            if (!bResetFound) timeVectComp.push_back(*iter);
        }

        lastMaster = 0;
        for (vector<master_time>::iterator iter = timeVectComp.begin(); iter != timeVectComp.end(); iter++) {
            // test if master time is strictly ascending
            if (lastMaster > iter->masterTime) {
                cout << "Incorrect master time detected: (" << iter->localTime <<  "," << iter->masterTime << ")" << endl;
            }
            lastMaster = iter->masterTime;
        }
    }

    cout << "Total time entries readed (after compensation): " << timeVectComp.size() << endl;

    //
    //  Read social data
    //
    fileName = argv[1];
    fileName += ".ee";
    ifstream fin(fileName.c_str(), ifstream::in );
    cout << "Input file: " << fileName << endl;
    vector<periodData*>  measurements;
    int     totalPeriodsReaded = 0;
    int     lineCount = 0;

    if (fin.is_open()) {
        bool    bIgnore = false;
        periodData* data = NULL;
        vector<master_time>::iterator iter = timeVectComp.begin();
        master_time closestMatch = {0, 0};
        if (iter != timeVectComp.end()) closestMatch = *iter;
        while (fin.getline(strIn, MAX_LINE_LEN)) {
            lineCount++;
            //cout << ".";
            line = strIn;
            if (line == "Received message...METADATA") {
                totalPeriodsReaded++;
                if (!bIgnore && (data != NULL)) measurements.push_back(data);
                if (bIgnore) delete data;

                data = new periodData;
                bIgnore = false;

                //Message: 0 Time: 10263 Period: 1 Responses: 4
                fin.getline(strIn, MAX_LINE_LEN);
                lineCount++;
                str = strstr(strIn, "Time: ");
                str += strlen("Time: ");
                data->localTime = atoi(str);
                str = strstr(str, "Period: ");
                str += strlen("Period: ");
                data->localPeriod = atoi(str);

                // Find closest matching entry in master time vector timeVectComp
                if (iter != timeVectComp.end()) {
                    while ((data->localTime > iter->localTime) && iter != timeVectComp.end()) iter++;
                    if (iter != timeVectComp.end()) closestMatch = *iter;
                }
                if (timeVectComp.size() > 0) {
                    // Deduce proper master time for given local time
                    data->globalTime = closestMatch.masterTime - (closestMatch.localTime - data->localTime);
                }
                else data->globalTime = 0;

                str = strstr(str, "Responses: ");
                str += strlen("Responses: ");
                int numResp = atoi(str);
                // when number of responses equal to 255, part of memory with no data stored is readed => ignore
                if (numResp == 255) {
                    bIgnore = true;
                }
                continue;
            }
            if (line == "Received message...DATA") {
                // do nothing
                continue;
            }
            if (strncmp(strIn, "Message: ", strlen("Message: ")) == 0) {
                //Message: 1 Friend: 44 RSSI: -68 Remaining responses: 3.
                str = strstr(strIn, "Friend: ");
                str += strlen("Friend: ");
                int nodeID = atoi(str);
                str = strstr(str, "RSSI: ");
                str += strlen("RSSI: ");
                int RSSI = atoi(str);

                if ((nodeID > nodeMappingLen) || (nodeMapping[nodeID] == -1)) { // test for too large or unused nodeIDs
                    cout << "Invalid nodeID detected: " << nodeID << " '" << strIn << "'" << endl;
                    bIgnore = true;
                }

                if (RSSI > 0 || RSSI < -100) {
                    cout << "Invalid RSSI detected: " << RSSI << " '" << strIn << "'" << endl;
                    bIgnore = true;
                }
                if (!bIgnore && (data != NULL)) {
                    // store only when valid nodeID was found
                    data->rssi[nodeMapping[nodeID]] = RSSI;
                }
                continue;
            }
        }
        fin.close();
        if (!bIgnore && (data != NULL)) measurements.push_back(data);
    }
    else {
        cout << "Fail to open input file" << endl;
    }


    cout << "Total periods readed: " << totalPeriodsReaded << endl;
    cout << "Valid periods readed: " << measurements.size() << endl;

    fileName = argv[1];
    fileName += ".dat";
    ofstream fout(fileName.c_str(), ifstream::out );
    cout << "Output file: " << fileName << endl;
    vector<periodData*>::iterator   iter;
    if (fout.is_open()) {
        for (iter = measurements.begin(); iter != measurements.end(); iter++) {
            fout << setw(7) << (*iter)->localPeriod << setw(10) << (*iter)->localTime << setw(10) << (*iter)->globalTime << "          ";
            for (int i = 0; i < nodeMappingInvLen; i++) fout << setw(5) << (*iter)->rssi[i];
            fout << endl;
            delete *iter;
        }

        fout.close();
    }
    else {
        cout << "Fail to open output file" << endl;
    }
    return 0;
}

